>* Learning to Fly: Let's simulate evolution in Rust! (pt 3) 译文（学习飞行：用 Rust 模拟种群进化 part3）
>* 原文链接：https://pwy.io/en/posts/learning-to-fly-pt3/
>* 原文作者：[Patryk27](https://github.com/Patryk27)
>* 译文来自：https://github.com/suhanyujie/article-transfer-rs/
>* 译者：[suhanyujie](https://github.com/suhanyujie)
>* 译者博客：[suhanyujie](https://ishenghuo.cnblogs.com/)
>* ps：水平有限，翻译不当之处，还请指正。
>* 标签：Rust, simulation,  genetic-algorithm, neural-network, rust, webassembly

This post is part of the learning-to-fly series:

1.[The Domain](https://pwy.io/en/posts/learning-to-fly-pt1/)
2.[The Neural Network](https://pwy.io/en/posts/learning-to-fly-pt1/)
3.[The Genetic Algorithm](https://pwy.io/en/posts/learning-to-fly-pt3/)
4.[The User Interface](https://pwy.io/en/posts/learning-to-fly-pt4/)

This is third part of the **Learning to Fly** series in which we’re coding a simulation of evolution using **neural network** and **genetic algorithm**:
> 使用**神经网络**和**遗传算法**对进化进行模拟，这是 **Learning to Fly** 系列的第三篇：

![](https://pwy.io/resources/learning-to-fly-pt1/intro-outcome.png)

_Figure 1. /en/projects/shorelark/_

In the previous post, we implemented a rudimentary feed-forward neural network that could propagate numbers through its randomized layers - it was the first milestone in our effors to create a functioning brain.
> 在上篇文章中，我们实现了一个初级的前馈神经网络，可以通过其随机层传播数字--这是我们努力创造一个正常运作的大脑神经网络的第一个里程碑。

Randomness can get us only so far, though - for the most part, evolution is about making small, incremental **changes** to the system, so that it gets better over time; so that our brains, from entirely haywire~ish, begin to gather knowledge and function as we expect from them (i.e. just catch the food, birdies!).
> 但随机性只能让我们走到这一步 -- 在大多数情况下，进化是基于现有系统进行小的、渐进式的**改变**，以便随着时间的推移变得更好；这样，我们的大脑就会从完全混乱的状态开始收集知识，并按照我们对它们的期望发挥作用（如，快抓住食物，小鸟们！）。

But how can we **train** a bunch of `Vec<f32>`?
> 但是，我们怎样才能**训练**一群 `Vec<f32>` 呢？

## Plan
Plan for today is straightforward: we’ll learn how a genetic algorithm works by implementing one in Rust. We’ll examine in-depth how [selection](https://en.wikipedia.org/wiki/Selection_(genetic_algorithm)), [crossover](https://en.wikipedia.org/wiki/Crossover_(genetic_algorithm)), and [mutation](https://en.wikipedia.org/wiki/Mutation_(genetic_algorithm)) all play together, allowing for a computer to find complex answers seemingly out of thin air.
> 现在目标很明确了：我们将通过 Rust 实现一个遗传算法来学习它是如何实现的。我们将深入研究怎样 [selection](https://en.wikipedia.org/wiki/Selection_(genetic_algorithm))，[crossover](https://en.wikipedia.org/wiki/Crossover_(genetic_algorithm)) 和 [mutation](https://en.wikipedia.org/wiki/Mutation_(genetic_algorithm)) 是如何一起发挥作用的，利用计算机找到确切的答案。

We’ll try to remain **generic**, meaning that instead of hard-coding a particular selection or crossover algorithm, we’ll use [traits](https://doc.rust-lang.org/book/ch10-02-traits.html) to create a versatile library that could be even published to [crates.io](https://crates.io/)!
> 我们将尽量保证**通用性**，这意味我们要利用 [traits](https://doc.rust-lang.org/book/ch10-02-traits.html) 创建一个通用的库，甚至可以发布到 [crates.io](https://crates.io/)！

As in the previous part, so today we’ll investigate various intricacies of Rust’s syntax, focusing a lot on the terminology.
> 如之前所说，我们会探究 Rust 语法的各种复杂之处，着重关注一下术语。

Hopefully, by the end of this post you’ll be able to say: I could have implemented it on my own!
> 希望在本文结束时，你敢说：你本可以自己实现它！

## Introduction

First, let’s recap how does a genetic algorithm work and what’s the point of this venture.
> 首先，让我们回顾一下遗传算法是如何工作的以及这个项目的重点是什么。

Our problem is that we’ve got an object - a neural network - that’s defined by a whole lot of parameters. There are so many of them that, even for the smallest of networks, we couldn’t possibly brute-force all of their combinations in our lifetime.
> 我们的问题是，我们有一个对象 —— 一个神经网络 —— 它是由一大堆参数定义的。它们的数量非常多，以至于即使是最小的网络，我们也不可能在有生之年对它们的所有组合进行暴力破解。

As you might remember, all the possible parameters in general are called a search space; so an erudite would say that our problem’s search space is huuuuge and then run away.
> 你们可能还记得，所有可能的参数通常被称为搜索空间；所以一个博学的人会说我们问题的搜索空间太大了，然后就走开了。

What we can do is to kinda mimic the nature: if we started with a bunch of random suboptimal solutions, we could try improving them to get - over time - gradually better answers.
> 我们能做的是模仿原生的东西：如果我们从一堆随机的次优解开始，我们可以尝试改进它们，随着时间的推移，逐渐得到更好的答案。

One of the methods that allows to simulate all this evolutionary machinery is the eponymous genetic algorithm: it starts with a bunch of random solution-candidates (a population), which are then improved using mutation and crossover, utilizing a fitness function to evaluate solutions (individuals) found so far:
> 一种能够模拟所有这些进化机制的方法是命名遗传算法：它从一堆随机的“解决方案”候选（一个种群）开始，然后使用突变和重组来改进它们，利用适应度函数来评估迄今为止已知的“解决方案”（个体）:

![](https://pwy.io/resources/learning-to-fly-pt3/intro-1.svg)

_Figure 2. Overview of the genetic algorithm; starting from top and and going clockwise: (1) estimating current solutions with a fitness function, (2) performing crossover and mutation, (3) starting the entire proces over on the new, improved population_

Because genetic algorithm involves working on random numbers, it’s an example of a probabilistic method.
> 因为遗传算法涉及到处理随机数，它是概率方法的一个例子。

Probabilistic algorithms trade accuracy for performance - they don’t always return the best answers, but usually get Pretty Close Pretty Cheaply ®.
> 概率算法以准确性换取性能 —— 它们不一定能得到最好的答案，但通常会得到非常接近且非常不错的结果。

Vegetable manufacturers don’t want you to know this, but there exists a straightforward procedure to becoming a carrot tycoon that’s actually based on a genetic algorithm:
> 蔬菜制造商不希望你知道这一点，但实际上有一个简单的程序，可以帮助你成为“胡萝卜大亨”，即基于遗传算法:

```
10  go to your garden
20  sow a few random carrots
30  wait for those carrots to sprout
40  choose the best carrot-children and sow them
50  goto 30

in this world:
- population = carrots
- individual = carrot
- mutation & crossover = happen automatically (free labor!)
- fitness function = your eyes & brain
```

By now, most of those words ought to sound familiar to you - we’ve already gone through the basics of evolutionary computation in the first article; by the end of this article, you’ll have answered questions such as:
> 到目前为止，这些词对你来说应该很熟悉了 —— 我们已经在第一篇文章中介绍了进化计算的基础知识；在本文结束时，你应该可以回答以下问题:

but how do you choose the individuals? there must be like a thousand ways to do it!
(psst: oh my yes, there are)
> 但是你如何选择个体呢？肯定有上千种方法可以做到!
> (psst:哦，我的天，是的)

but how do you represent their genomes? there must be like a thousand ways to do it!
(psst: oh my yes, there are)
> 如何表示它们的基因组呢？肯定有上千种方法可以做到!
> (psst:哦，我的天，是的)

but how do you implement it in Rust? you promised it’ll work inside a web browser!
(psst: oh my yes, it will)
> 如何用 Rust 实现它呢？你承诺过它可以运行在网络浏览器中!
> (psst:哦，我的天，它会的)

## Coding: Outline

We’ll start by creating a second crate inside our workspace:
> 我们现在要在工作区中创建第二个 crate：

```
$ cd shorelark/libs
$ cargo new genetic-algorithm --lib
```

Oh, that’s some nice genetic-algorithm/src/lib.rs Cargo created for us in there - let’s replace it with a just as good entry point:
> 这是一个很不错的遗传算法库 —— `algorithm/src/lib.rs` Cargo 

```rust
pub struct GeneticAlgorithm;

impl GeneticAlgorithm {
    pub fn new() -> Self {
        Self
    }
}
```

Our genetic algorithm, as all good objects do, will provide only one functionality - sometimes it’s called iterate, sometimes it’s called step or process; I’ve tossed a coin and decided on:

```rust
impl GeneticAlgorithm {
    /* ... */

    pub fn evolve(&self) {
        todo!()
    }
}
```

What are we evolving? A population, of course!

```rust
impl GeneticAlgorithm {
    pub fn evolve(&self, population: &[???]) -> Vec<???> {
        todo!()
    }
}
```

Our actual problem will depend on neural networks, but since we want for this library to be generic, we can’t force it to accept a hard-coded NeuralNetwork - instead, we can introduce a type parameter:

```rust
impl GeneticAlgorithm {
    pub fn evolve<I>(&self, population: &[I]) -> Vec<I> {
        todo!()
    }
}
```

I stands for individual.

As for the Rust’s terminology:

// visibility  generics   _ function parameters
// |          _|     ____|  (or just "parameters")
// |         |      |
// v-v       v-----vv----------v
   pub fn foo<'a, T>(bar: &'a T) { /* ... */ }
//            ^^  ^  ^--------^
//            |   |  |
//            |   |  function parameter
//            |   |  (or just "parameter")
//            |   type parameter
//            lifetime parameter
(https://doc.rust-lang.org/reference/items/functions.html)

If you wanted to read this signature aloud, you’d say:

public function foo is generic over lifetime a and type T, and it accepts a single parameter named bar that is a reference to T.

That was function’s definition - on the other hand, the place where you invoke a function is named call site and the values you specify there are called arguments:

// v-----------------------v call site
   foo::<'static, f32>(&1.0);
//       ^-----^  ^-^  ^--^
//       |        |    |
//       |        |    function argument
//       |        |    (or just "argument")
//       |        type argument
//       lifetime argument
Most of this vernacular (e.g. the difference between argument and parameter) is universal across all the programming languages, so it’s worth remembering.

Learning from past mistakes, let’s not forget about preconditions:

```rust
pub fn evolve<I>(&self, population: &[I]) -> Vec<I> {
    assert!(!population.is_empty());

    /* ... */
}
```

As for the algorithm itself - the outline is:

```rust
pub fn evolve<I>(&self, population: &[I]) -> Vec<I> {
    /* ... */

    (0..population.len())
        .map(|_| {
            // TODO selection
            // TODO crossover
            // TODO mutation
            todo!()
        })
        .collect()
}
```

## Coding: Selection

At this point, inside the loop, we have to pick two individuals - they will become parents and "create" us a digital offspring.

_(imagining that chip made of sand can create "offspring" hits my uncanny valley right in the feels.)_

Choosing individuals is called the selection stage of a genetic algorithm, and it should satisfy following two properties:

each individual should have a non-zero chance of being picked,

an individual with a higher fitness score should get picked, on average, more often than an individual with a lower fitness score.

Because we’ll have to know fitness scores, let’s start off by thinking how we want users to specify their fitness function; as trivial as it sounds, there are at least two exceptive approaches we can apply:

Fitness function as a parameter:

```rust
pub fn evolve<I>(
    &self,
    population: &[I],
    evaluate_fitness: &dyn Fn(&I) -> f32,
) -> Vec<I> {
    /* ... */
}
```

Fitness score as a property of an individual:

```rust
pub trait Individual {
    fn fitness(&self) -> f32;
}

pub fn evolve<I>(&self, population: &[I]) -> Vec<I>
where
    I: Individual,
{
    /* ... */
}
```

First approach:

* ✅ allows to provide many different fitness functions for one kind of an individual, which might prove to be useful for somebody (not us, though),

* ❌ requires specifying fitness function for each invocation of .evolve(), which feels a bit shoehorned.

Second approach:

* ✅ allows to encapsulate all the individual-oriented attributes into a single trait, making it easy for users to discover what they need to provide,

* ❌ specifying different fitness functions is possible, but a bit more tricky (can you figure out how and why?).

My guts vote 133:7 for introducing a trait (a trait that, as you’ll see later, we would need anyway), so a trait it is.

As for the selection method, we’ll use an algorithm called fitness proportionate selection (also known as roulette wheel selection), as it’s easy to reason about; to understand how it works, let’s imagine we have following three individuals:

Individual | Fitness score| Fitness score %
|:---- |:---|:---- |
| A | 3 | `3 / (1 + 2 + 3) = 3 / 6 = 50%` |
| B | 2 | `2 / (1 + 2 + 3) = 2 / 6 = 33%` |
| C | 1 | `1 / (1 + 2 + 3) = 1 / 6 = 16%` |

If we placed them all on a roulette wheel - or a pie chart, for all it matters - with each individual getting a slice of wheel as large as proportion of their fitness score to the rest:

![](https://pwy.io/resources/learning-to-fly-pt3/coding-selection-1.svg)

_Figure 3. A pie chart - or a roulette wheel, if you squeeze your eyes enough - illustrating individuals from the table above_

... randomizing an individual would boil down to "spinning" the wheel with random force and seeing what came up:

![](https://pwy.io/resources/learning-to-fly-pt3/coding-selection-2.svg)

In practice, fitness proportionate selection is rather frowned upon - it’s because it allows for the best individuals to dominate the simulation.

Say, your genetic algorithm finds a solution that’s exponentially better than the rest:

![](https://pwy.io/resources/learning-to-fly-pt3/coding-selection-3.svg)

coding selection 3
... when it happens, fitness proportionate selection will happily choose this green solution 99% of the time, making rest of the individuals an army of copy-pasted, green clones.

You might think:

isn’t finding the best solution, like, the whole point?

... it is; but it’s important to remember that solutions found by genetic algorithm are always best to date - if you discard a seemingly unpromising candidate too early, you will never know whether tuning some parameter wouldn’t make it an even better solution in the long run.

To put it another way:

the more diverse humans you have, the greater chance one of them is a trombone prodigy

For simplicity, we’ll continue with the roulette wheel selection - but should you feel frisky, I’ll just say that rank selection is example of an algorithm that doesn’t exhibit this dominating behavior, and it will work with our birdies, too!


Living up to the generic-ness promise, instead of hard-coding our library to always use roulette wheel selection, let’s create a trait - this way users will be able to use any algorithm they fancy:

```rust
pub trait SelectionMethod {
    fn select(&self);
}
```

A selection method has to have access to the entire population:

```rust
pub trait SelectionMethod {
    fn select<I>(&self, population: &[I]) -> &I
    where
        I: Individual;
}
```

For clarity, let’s annotate the output’s lifetime:

```rust
pub trait SelectionMethod {
    fn select<'a, I>(&self, population: &'a [I]) -> &'a I
    where
        I: Individual;
}
```

We’re going to need random numbers any minute now, so let’s add rand to libs/genetic-algorithm/Cargo.toml:

```toml
# ...

[dependencies]
rand = "0.8"

[dev-dependencies]
rand_chacha = "0.3"
```

>Each crate in a workspace has its own set of dependencies - the rand we’ve previously added to libs/neural-network/Cargo.toml is not automatically visible in any other crate in the workspace.

Learning from our past troubles with thread_rng(), let’s already pass the pseudo-random number generator via a parameter:

```rust
use rand::RngCore;

pub trait SelectionMethod {
    fn select<'a, I>(
       &self,
       rng: &mut dyn RngCore,
       population: &'a [I],
    ) -> &'a I
    where
        I: Individual;
}
```

Ain’t that a beautiful signature?

---

You might be wondering why we don’t take a step further and make select() generic over the PRNG, too:

```rust
pub trait SelectionMethod {
    fn select<'a, R, I>(
       &self,
       rng: &mut R,
       population: &'a [I],
    ) -> &'a I
    where
        R: RngCore,
        I: Individual;
}
```

First, let’s catch up on the vernacular - in general:

* dyn Trait, &dyn Trait and &mut dyn Trait imply dynamic dispatch,

* T, &T and &mut T imply static dispatch.

Dispatching is the way compiler answers the question "which method should get called here, exactly?" for generic types:

```rust
fn foo() {
   bar();

   // ^ compiling this call is easy, because it always transfers
   // control into `bar`
}

fn bar() {
   println!("yas queen");
}

fn method(obj: &dyn SomeTrait) {
    obj.method();

    // ^ compiling this call is harder, because there's no single
    // function this could refer to - each implementation of the
    // trait might provide its own `fn method()`
    //
    // in general, this is called *polymorphism*
}
```

For the sake of an example, let’s consider this trait with its two implementations:

```rust
trait Animal {
    fn kind(&self) -> &'static str;
}

// --

struct Chinchilla;

impl Animal for Chinchilla {
    fn kind(&self) -> &'static str {
        "chinchilla"
    }
}

// --

struct Viscacha;

impl Animal for Viscacha {
    fn kind(&self) -> &'static str {
        "viscacha"
    }
}
```

If you wanted to create a function that prints kind of any animal, you could do it twofold:

open in playground

```rust
// Uses static dispatch
// (aka static polymorphism)
fn print_kind_static<A>(animal: &A)
    where A: Animal
{
    println!("{}", animal.kind());
}

// Uses dynamic dispatch
// (aka dynamic polymorphism, aka runtime polymorphism)
fn print_kind_dynamic(animal: &dyn Animal) {
    println!("{}", animal.kind());
}

fn main() {
    print_kind_static(&Chinchilla);
    print_kind_static(&Viscacha);

    print_kind_dynamic(&Chinchilla);
    print_kind_dynamic(&Viscacha);
}
```

From a distance, both function look alike - what’s the difference, then?

print_kind_static() uses a technique called monomorphization - meaning that for each Animal this function is invoked with, compiler transparently generates a dedicated, "copy-pasted" version of that function:

```rust
fn print_kind_static__chinchilla(animal: &Chinchilla) {
    println!("{}", Chinchilla::kind(animal));
}

fn print_kind_static__viscacha(animal: &Viscacha) {
    println!("{}", Viscacha::kind(animal));
}

fn main() {
    print_kind_static__chinchilla(&Chinchilla);
    print_kind_static__viscacha(&Viscacha);
}
```

At this point you can see why it’s called static dispatch - underneath, compiler replaces dynamic traits with static types.

Monomorphization has a drawback of being a bit slower to compile (instead of just one function, compiler has to process many of them), but usually it results in a faster, more optimized code at runtime; it can make a noticeable difference for applications that call such generic functions, say, million times per second.

print_kind_dynamic(), on the other hand, uses a technique called vtable ("virtual table"), where each implementation is created a dedicated "proxy table" that maps to concrete functions:

```rust
// This is pseudo-Rust, just to show the concept

struct AnimalVtable {
    // Pointer to a concrete implementation of the `kind()` method
    kind: fn(*const ()) -> &'static str,
}

const CHINCHILLA_VTABLE: AnimalVtable = AnimalVtable {
    kind: Chinchilla::kind,
};

const VISCACHA_VTABLE: AnimalVtable = AnimalVtable {
    kind: Viscacha::kind,
};

fn print_kind_dynamic(
    animal_obj: *const(),
    animal_vtable: &AnimalVtable,
) {
    println!("{}", animal_vtable.kind(animal_obj));
}

fn main() {
    print_kind_dynamic(&Chinchilla, &CHINCHILLA_VTABLE);
    print_kind_dynamic(&Viscacha, &VISCACHA_VTABLE);
}
```

Since all implementations can be described via AnimalVtable, print_kind_dynamic() doesn’t have to be monomorphized - depending on the underlying type, compiler will simply pass different vtable.

In this case, the drawback is that each time you call print_kind_dynamic(), it has to go through this additional "proxy table", which makes it theoretically slower than print_kind_static(); more often than not the difference is not meaningful, though.

Circling back to the original question: so why not where R: RngCore?

Because we won’t be invoking this PRNG million times per second, and so the additional maintenance burden is - to me - a cake not worth the candle.

As for the implementation, we could do it by hand:

```rust
use rand::Rng;
use rand::seq::SliceRandom;

pub struct RouletteWheelSelection;

impl RouletteWheelSelection {
    pub fn new() -> Self {
        Self
    }
}

impl SelectionMethod for RouletteWheelSelection {
    fn select<'a, I>(
       &self,
       rng: &mut dyn RngCore,
       population: &'a [I],
    ) -> &'a I
    where
        I: Individual,
    {
        assert!(!population.is_empty());

        let total_fitness: f32 = population
            .iter()
            .map(|individual| individual.fitness())
            .sum();

        // This is a naïve approach for demonstration purposes; a more
        // efficient implementation could invoke `rng` just once
        loop {
            let indiv = population
                .choose(rng)
                .expect("got an empty population");

            let indiv_share = indiv.fitness() / total_fitness;

            if rng.gen_bool(indiv_share as f64) {
                return indiv;
            }
        }
    }
}
```

... but a code par excellence would be <drums/> no code at all!

If you go through rand-'s documentation, you might just spot a trait called SliceRandom; if you take a look inside it, you might just spot a method called choose_weighted() that happens to be doing exactly the thing we need:

```rust
impl SelectionMethod for RouletteWheelSelection {
    fn select<'a, I>(
       &self,
       rng: &mut dyn RngCore,
       population: &'a [I],
    ) -> &'a I
    where
        I: Individual,
    {
        population
            .choose_weighted(rng, |individual| individual.fitness())
            .expect("got an empty population")
    }
}
```

_(thanks to @javiertury for pointing out that this method exists.)_

Apart from trusting rand-s developers, how can we be sure choose_weighted() does the thing we need? By testing it!

```rust
#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test() {
        todo!();
    }
}
```

Path to the TDD-nirvana is a path strawn with roses, and we’re about to get bit by one of their thorns:

```rust
use rand::SeedableRng;
use rand_chacha::ChaCha8Rng;

#[test]
fn test() {
    let mut rng = ChaCha8Rng::from_seed(Default::default());

    let population = vec![ /* what here? */ ];

    let actual = RouletteWheelSelection::new()
        .select(&mut rng, &population);

    assert!(/* what here? */);
}
```

At this point we’ve got two problems:

* Since Individual is a trait, how can we fake it for testing purposes?

* Since .select() returns just a single individual, how can we be sure it’s random?
(obligatory xkcd.)

Starting from the top: creating fake objects for testing purposes is called mocking - and while there are mocking solutions for Rust, I gotta admit I’ve never been a fan of the concept of a mock, as best presented in a song:

Friend, either you're closing your eyes
To a situation you do not wish to acknowledge
Or you are not aware of the caliber of disaster indicated
By the presence of mocks in your repository

[...]

Oh yes we got trouble, trouble, trouble!
With a "T"! Gotta rhyme it with "M"!
And that stands for Mock

_(but for real, my contempt for mocks deserves its own post.)_

My suggestion - that doesn’t require any external crates - it to create a dedicated testing-struct:

```rust
#[cfg(test)]
#[derive(Clone, Debug)]
pub struct TestIndividual {
    fitness: f32,
}

#[cfg(test)]
impl TestIndividual {
    pub fn new(fitness: f32) -> Self {
        Self { fitness }
    }
}

#[cfg(test)]
impl Individual for TestIndividual {
    fn fitness(&self) -> f32 {
        self.fitness
    }
}
```

... which can be then used as:

```rust
#[test]
fn test() {
    let population = vec![
        TestIndividual::new(2.0),
        TestIndividual::new(1.0),
        TestIndividual::new(4.0),
        TestIndividual::new(3.0),
    ];

    /* ... */
}
```

What about the assertion, though? A test such as this one:

```rust
#[test]
fn test() {
    /* ... */

    let actual = RouletteWheelSelection::new()
        .select(&mut rng, &population);

    assert!(actual, &population[2]);
}
```

... doesn’t inspire confidence, as it doesn’t prove that fitness scores are being considered; a hypothetical, invalid implementation:

```rust
impl SelectionMethod for RouletteWheelSelection {
    fn select<'a, I>(/* ... */) -> &'a I
    where
        I: Individual,
    {
        &population[2]
    }
}
```

... would pass such test with flying colors!

Fortunately, we are not doomed - since we want to assess probability, instead of invoking .select() once, we can do it many times and look at the histogram:

![](https://pwy.io/resources/learning-to-fly-pt3/coding-selection-4.svg)

_Figure 4. A histogram - X axis represents items, Y axis represents frequency; also, to make hay while the sun shines, I hereby dub this type of histogram the The Johnny Bravo Chart_

```rust
use std::iter::FromIterator;

#[test]
fn test() {
    let method = RouletteWheelSelection::new();
    let mut rng = ChaCha8Rng::from_seed(Default::default());

    let population = vec![
        TestIndividual::new(2.0),
        TestIndividual::new(1.0),
        TestIndividual::new(4.0),
        TestIndividual::new(3.0),
    ];

    let mut actual_histogram = BTreeMap::new();

    //               there is nothing special about this thousand;
    //          v--v a number as low as fifty might do the trick, too
    for _ in 0..1000 {
        let fitness = method
            .select(&mut rng, &population)
            .fitness() as i32;

        *actual_histogram
            .entry(fitness)
            .or_insert(0) += 1;
    }

    let expected_histogram = BTreeMap::from_iter(vec![
        // (fitness, how many times this fitness has been chosen)
        (1, 0),
        (2, 0),
        (3, 0),
        (4, 0),
    ]);

    assert_eq!(actual_histogram, expected_histogram);
}
```

---
Notice that while building the histogram, we’re casting fitness scores from f32 to i32:

```rust
let fitness = method
    .select(&mut rng, &population)
    .fitness() as i32;
```

We have to do that, because floating-point numbers in Rust don’t implement the Ord trait, making it impossible to use them as a BTreeMap-'s key:

```rust
use std::collections::BTreeMap;

fn main() {
    let mut map = BTreeMap::new();
    map.insert(1.0, "one point zero");
}
```

[open in playground](https://play.rust-lang.org/?version=nightly&mode=debug&edition=2018&gist=535648a70ce2e61f12b8c38021fddbef)

```
error[E0277]: the trait bound `{float}: Ord` is not satisfied
  |
  |     map.insert(1.0, "one point zero");
  |         ^^^^^^ the trait `Ord` is not implemented for `{float}`
```

The reason is that floating-point numbers, as defined by the IEEE 754 standard, are not a totally ordered set - namely, comparing NaN-s is problematic, because:

```
NaN != NaN
```

Practically, it means that had you ever inserted a NaN into a map, not only would you be unable to retrieve that particular entry using .get(), but you could break BTreeMap-'s internal structure, making it impossible to retrieve any item.

(by the way, that’s also true for custom implementations of Ord and PartialOrd - if they don’t satisfy asymmetry and transitivity, you’re gonna have a bad time.)

If you feel like exploring this topic more:

* https://doc.rust-lang.org/std/cmp/trait.PartialOrd.html

* https://www.reddit.com/r/rust/comments/cbrpll/why_doesnt_stdcmpord_is_implemented_for_floats/

* https://www.reddit.com/r/rust/comments/29kia3/no_ord_for_f32/

* https://stackoverflow.com/questions/26489701/why-does-rust-not-implement-total-ordering-via-the-ord-trait-for-f64-and-f32
---

cargo test (or cargo test --workspace, if you’re in the virtual manifest’s directory) returns:

```
thread '...' panicked at 'assertion failed: `(left == right)`
  left: `{1: 98, 2: 202, 3: 278, 4: 422}`,
 right: `{1: 0, 2: 0, 3: 0, 4: 0}`'
```

... proving that choose_weighted() does work as advertised (higher fitness scores were chosen more frequently), so let’s fix the test:

```rust
#[test]
fn test() {
    /* ... */

    let expected_histogram = BTreeMap::from_iter(vec![
        // (fitness, how many times this fitness has been chosen)
        (1, 98),
        (2, 202),
        (3, 278),
        (4, 422),
    ]);

    /* ... */
}
```

Voilà - we’ve tested the untestable!

Before moving on, let’s take a moment to ponder the nature of existence - and maybe rustify our code a little bit as there are some things we could get improved.

First, constructing maps via ::from_iter() is kinda messy - not only you have to create a vector on the way, but you’re limited to (key, value) tuples that look foreign to an untrained eye.

As always when it comes to Rust’s syntax being overly verbose: just macro it ™; in case we’ll make use of a crate named maplit:

```toml
# ...

[dev-dependencies]
# ...
maplit = "1.0"
```

... that provides a handy macro called btreemap!:

```rust
#[test]
fn test() {
    /* ... */

    let expected_histogram = maplit::btreemap! {
        // fitness => how many times this fitness has been chosen
        1 => 98,
        2 => 202,
        3 => 278,
        4 => 422,
    };

    /* ... */
}
```

Moreover, we can use Iterator::fold() to simplify the loop:

```
#[test]
fn test() {
    /* ... */

    let actual_histogram: BTreeMap<i32, _> = (0..1000)
        .map(|_| method.select(&mut rng, &population))
        .fold(Default::default(), |mut histogram, individual| {
            *histogram
                .entry(individual.fitness() as _)
                .or_default() += 1;

            histogram
        });

    /* ... */
}
```

as _ means "compiler, pretty please infer what type is required and cast this value into it".

While this operation feels pointless, it performs the same function it did a few code blocks above (i.e. converting f32 to i32) - the only difference is that now that we’ve explicitly stated our map’s key using : BTreeMap<i32, _>, we don’t have to do it inside the casting.

(it wouldn’t be wrong to repeat the type and write as i32 - it’s just my preference to avoid duplicating type annotations.)

Another way we could’ve written this code is:

```rust
let actual_histogram = (0..1000)
    .map(|_| method.select(&mut rng, &population))
    .fold(BTreeMap::default(), |mut histogram, individual| {
        *histogram
            .entry(individual.fitness() as i32)
            .or_default() += 1;

        histogram
    });
```

All those three approaches (including the initial one with an explicit loop) are equally valid Rust code - use whichever one you find the most readable.

Having the selection algorithm ready, let’s recap where we stopped:

```rust
pub fn evolve<I>(&self, population: &[I]) -> Vec<I>
where
    I: Individual,
{
    /* ... */

    (0..population.len())
        .map(|_| {
            // TODO selection
            // TODO crossover
            // TODO mutation
            todo!()
        })
        .collect()
}
```

What we have to figure out now is how to pass SelectionMethod there - I see two approaches:

* 1.Using parameter:

```rust
pub fn evolve<I, S>(
    &self,
    population: &[I],
    selection_method: &S,
) -> Vec<I>
where
    I: Individual,
    S: SelectionMethod,
{
    /* ... */
}
```

* 2.Using constructor:

```rust
pub struct GeneticAlgorithm<S> {
    selection_method: S,
}

impl<S> GeneticAlgorithm<S>
where
    S: SelectionMethod,
{
    pub fn new(selection_method: S) -> Self {
        Self { selection_method }
    }

    pub fn evolve<I, S>(&self, population: &[I]) -> Vec<I>
    where
        I: Individual,
    {
        /* ... */
    }
}
```

When I’m facing this kind of decision, I think how often users will need to affect that particular value: a population is usually different each time someone calls .evolve(), so it’s convenient to accept it via parameter; on the other hand, selection algorithm generally remains identical for the whole simulation, so it’ll be wiser to go with constructor.

Now we’re almost ready to invoke the selection method:

```rust
pub fn evolve<I>(&self, population: &[I]) -> Vec<I>
where
    I: Individual,
{
    /* ... */

    (0..population.len())
        .map(|_| {
            let parent_a = self
                .selection_method
                .select(population);

            let parent_b = self
                .selection_method
                .select(population);

            // TODO crossover
            // TODO mutation
            todo!()
        })
        .collect()
}
```

... the only thing we’re missing is the PRNG:

```rust
pub fn evolve<I>(
    &self,
    rng: &mut dyn RngCore,
    population: &[I],
) -> Vec<I>
where
    I: Individual,
{
    /* ... */

    (0..population.len())
        .map(|_| {
            let parent_a = self
                .selection_method
                .select(rng, population);

            let parent_b = self
                .selection_method
                .select(rng, population);

            // TODO crossover
            // TODO mutation
            todo!()
        })
        .collect()
}
```

---
You might be wondering why rng is passed via .evolve() instead of going through the constructor - surely the random number generator doesn’t change that frequently!

Well, this decision was more subtle - to understand why, let’s explore other ways we could have written that code:

By accepting an owned PRNG via the constructor:

```rust
pub struct GeneticAlgorithm<R> {
    rng: R,
}

impl<R> GeneticAlgorithm<R>
where
    R: RngCore,
{
    pub fn new(rng: R) -> Self {
        Self { rng }
    }
}
```

By accepting a borrowed PRNG via the constructor:

```rust
pub struct GeneticAlgorithm<'r> {
    rng: &'mut dyn RngCore,
}

impl<'r> GeneticAlgorithm<'r> {
    pub fn new(rng: &'r mut dyn RngCore) -> Self {
        Self { rng }
    }
}
```

The first approach is something I would suggest had we been implementing our simulation in C# or Java - it’s a different story in Rust, because if we move rng inside the constructor, we won’t be able to use it in other places of the application:

```rust
fn main() {
    let rng = /* ... */;
    let ga = GeneticAlgorithm::new(rng);

    // oh no, we can't use this `rng` anymore!
    if rng.gen_bool() {
        /* ... */
    } else {
        /* ... */
    }
}
```

You could argue the same already happens for SelectionMethod:

```rust
fn main() {
    let sp = RouletteWheelSelection::new();
    let ga = GeneticAlgorithm::new(sp);

    // oh no, we can't use this `sp` anymore!
    if sp.something() {
        /* ... */
    }
}
```

... but in my opinion there’s a difference in that Rng is a more versatile trait - it makes sense to use it outside the GeneticAlgorithm, which cannot be said about SelectionMethod.

All said, you’d be right to call this a far-fetched explanation - picking a "more versatile" trait is nothing but a gut feeling; if you ask me, the owned-value approach is correct, just slightly inferior in this particular case.

As for the &mut dyn RngCore variant, I consider it the worst - it requires a unique borrow (&mut) on rng, so not only it "locks" PRNG for the lifetime of the genetic algorithm:

```rust
fn main() {
    let rng = /* ... */;
    let ga = GeneticAlgorithm::new(&mut rng);

    // oh no, we still can't use this `rng`!
    let population = if rng.gen_bool() {
        /* ... */
    } else {
        /* ... */
    };

    ga.evolve(population);
}
```

... but it also prevents otherwise valid use cases such as this one:

```rust
struct Simulation {
    rng: ChaCha8Rng,
    ga: GeneticAlgoritm<'whats_this_lifetime??>,
}
```

By the way, this is called a self-referential struct - without referring to eldritch unsafe magicks, it cannot be declared in Rust.

More on this topic, if you feel curious:

* https://stackoverflow.com/questions/32300132/why-cant-i-store-a-value-and-a-reference-to-that-value-in-the-same-struct

* https://boats.gitlab.io/blog/post/2018-01-25-async-i-self-referential-structs/

---


## Coding: Crossover

Now that we have chosen two parent-individuals, it’s time for the crossover phase.

Crossover (also known as recombination) takes two individuals and "mixes" them, yielding a new~ish solution:

![coding crossover 1](https://pwy.io/resources/learning-to-fly-pt3/coding-crossover-1.svg)

Compared to randomizing brand new individuals, crossover is neat in the sense that it tries to preserve knowledge - the idea is that combining two solutions will usually yield a solution that’s both new and at least as good as the two solutions we already have; this allows to explore the search space without the risk of losing the best solutions discovered so far.

As in real world, crossover doesn’t actually happen on an individual, but rather on their chromosomes - which is a fancy word for "an encoding of a solution":

![](https://pwy.io/resources/learning-to-fly-pt3/coding-crossover-2.svg)

A chromosome (also called genotype, though I’m convinced a biologist dies each time someone juxtaposes both terms) is usually built from genes, and its structure depends on the underlying problem - sometimes it’s convenient to model a chromosome as a bitset:

![coding crossover 3](https://pwy.io/resources/learning-to-fly-pt3/coding-crossover-3.svg)

... sometimes it’s more convenient to have a string:

![](https://pwy.io/resources/learning-to-fly-pt3/coding-crossover-4.svg)

... and we’ll re-use what we already have - a bunch of floating-point numbers (i.e. weights of the neural network):

![](https://pwy.io/resources/learning-to-fly-pt3/coding-crossover-5.svg)

As for the code:

```rust
#[derive(Clone, Debug)]
pub struct Chromosome {
    genes: Vec<f32>,
}
```

Instead of exposing genes directly (via pub genes: …​), we’ll provide a handful of functions allowing to peek inside the chromosome:

```rust
impl Chromosome {
    pub fn len(&self) -> usize {
        self.genes.len()
    }

    pub fn iter(&self) -> impl Iterator<Item = &f32> {
        self.genes.iter()
    }

    pub fn iter_mut(&mut self) -> impl Iterator<Item = &mut f32> {
        self.genes.iter_mut()
    }
}
```

If you’re assiduous, you might even write a handful of tests for them; some people would consider testing such tiny functions overzealous, but I say we go for it:

```rust
#[cfg(test)]
mod tests {
    use super::*;

    fn chromosome() -> Chromosome {
        Chromosome {
            genes: vec![3.0, 1.0, 2.0],
        }
    }

    mod len {
        use super::*;

        #[test]
        fn test() {
            assert_eq!(chromosome().len(), 3);
        }
    }

    mod iter {
        use super::*;

        #[test]
        fn test() {
            let chromosome = chromosome();
            let genes: Vec<_> = chromosome.iter().collect();

            assert_eq!(genes.len(), 3);
            assert_eq!(genes[0], &3.0);
            assert_eq!(genes[1], &1.0);
            assert_eq!(genes[2], &2.0);
        }
    }

    mod iter_mut {
        use super::*;

        #[test]
        fn test() {
            let mut chromosome = chromosome();

            chromosome.iter_mut().for_each(|gene| {
                *gene *= 10.0;
            });

            let genes: Vec<_> = chromosome.iter().collect();

            assert_eq!(genes.len(), 3);
            assert_eq!(genes[0], &30.0);
            assert_eq!(genes[1], &10.0);
            assert_eq!(genes[2], &20.0);
        }
    }
}
```

Seizing the day, let’s catch up on some cool traits from the standard library:

* 1.There’s Index, which allows to use the indexing operator - [] - on your type:
