>* Hecto: Build your own text editor in Rust 译文（用 Rust 构建自己的文本编辑器）
>* 原文链接：https://www.philippflenker.com/hecto/
>* 原文作者：[Philipp Flenker](https://github.com/pflenker)
>* 译文来自：https://github.com/suhanyujie/article-transfer-rs/
>* 译者：[suhanyujie](https://ishenghuo.cnblogs.com/)
>* ps：水平有限，如有不当之处，欢迎指正
>* 标签：Rust，text-editor

![](https://www.philippflenker.com/assets/2019-11-08-hecto-chapter-7.png)

这是一系列展示如何在 Rust 中构建文本编辑器的博客文章。它是 Rust 中 kilo 的重新实现，正如本教程中所概述的。与原来的小册子一样，这些博客文章指导您完成构建基本文本编辑器 hecto 的所有步骤。

通过应用更改、保存和运行程序，您几乎总是能够看到您的更改正在发生。我会尽可能详细地解释每一个步骤，有时会非常详细，并经常链接到其他页面。请随意浏览资料文章或者忽略链接，仅仅通过应用代码修改和观察您的文本编辑器的逐步完善，就可以学到很多东西！

## 为什么撰写这个系列？
我一直认为，每个软件工程师至少需要具备两种编程语言的基本知识。然而，我不得不承认，在过去的几年里，除了 JavaScript，我对几乎所有东西的了解都开始减少了。

这就是为什么我开始学习 Rust，我重新实现了 `kilo` 作为一种学习经验。但是为什么呢？为了学习 Rust，我想重新实现一个易于理解的软件，这样我就可以专注于语言而不会迷失在实现细节中。但是我不想重新实现我用 JavaScript 实现过的东西，因为我认为 JavaScript 是为不同于 Rust 的场景而设计的。或者换句话说: 如果你是一个水管工，你最好学会如何使用斧头，用它来砍倒一些树木，而不是用斧头疏通水槽。

`kilo` 的复杂程度足以构成一个挑战，当我读到它的时候，我希望可以使用 Rust 实现它 —— 现在它可以了！

为什么取这个名字？`hecto` 追求的目标比 `kilo` 更适中。它的目标不是小型化，它甚至不是我自己的想法 —— 所以给它一个比它的前辈更加谦逊的名字似乎是合适的。

## License
虽然这些博客文章参考这个[原始教程](https://viewsourcecode.org/snaptoken/kilo/index.html)，但代码已经用 Rust 实现，不仅通过调用最接近的“Rust 对应函数”，而且通过尝试解决问题的“Rust 方式”。同样，在 Rust 的上下文中，所有的内容都被检查和修改，并且在许多情况下被重写。因此，本教程应被视为原始 C 版本的“Rust 混合体”。

## 反馈
我很高兴你读了我的作品，也很想[听听你的意见](https://www.philippflenker.com/about) —— 尤其是当你遇到困难或者找到了解决具体问题的更好方法的时候。请记住，这主要是一个练习，让我了解 Rust，所以如果有一个更好的方法来做事情，请[告诉我](https://www.philippflenker.com/about)！

## 2022 年最新更新
我写这个教程已经有好几年了，从那以后 Rust 发生了很多变化。这意味着本教程中描述的代码可能不再有效，它可能会产生在我第一次编写本教程时没有出现的警告或错误。人们仍然在向我强调 bug，或者用 pr 来修复这些问题 —— 然而，我没有太多精力来更新本教程中的所有代码示例。只要有意义，我至少会用特别区域和提示更新教程的文本，以便通过一些努力，您应该能够完成教程。

## 系列第 1 篇 —— 设置

哈，第一步。你难道不喜欢从零开始吗？然后选择一些独一无二的砖头，你会在上面建造你的宫殿吗？

不幸的是，在构建计算机程序时，第一步可能会变得... 很复杂。也令人沮丧。您必须确保您的环境是适合 Rust 的，并且您必须弄清楚如何在该环境中编译和运行您的程序。

幸运的是，使用 Rust 设置一个开发环境是比较众所周知的，除了文本编辑器、Rust 和 cargo 之外，我们不需要其他任何东西。要安装这些程序，我们将使用一个名为 `rustup` 的程序，但是还有其他可用的方法来启动和运行 Rust。

如果您正在使用 Linux，或任何 Unix 系统，包括 MacOS，只要按照下面的说明。如果你在 Windows 上，你需要访问一个 Linux 系统。最简单的方法是使用 Windows 的 Linux 子系统（WSL），可以在 Windows 10 上使用。您也可以在 Windows 上直接运行和安装 Rust，但是在遵循本教程的过程中，您很快就会遇到问题，并且可能无法完成本教程。

> 自从本教程在 2019 年发布以来，发生了很多变化，其中一个变化是，对于本教程中的一个库，现在存在一个同样运行在 Windows 计算机上的替代方案。一旦我们进展到那一步，我将指出替代方案，但本教程仍将使用旧的库。

## 如何用 rustup 安装 rust

如果您访问 [rustup 网站](https://rustup.rs/)，它会尝试自动检测您的操作系统，并显示安装 rustup 的最佳方式。通常，您下载并执行一个脚本 `rustup-init`，它为您执行安装。

但是，如果你觉得下载和执行远程脚本对您来说不方便，那么您可以单击[其他安装选项](https://github.com/rust-lang/rustup.rs/#other-installation-methods)并直接下载 `rust-init`，以便您的平台开始安装。

确保环境符合 Linux、 macOS 或 Windows 子系统上的 Linux (WSL)

安装完成后，安装程序会通过打印告诉你:

```
Rust is installed now. Great!
```

要开始使用 Rust，您要么需要重新启动终端，要么需要输入

```Shell
$ source $HOME/.cargo/env
```

## 添加链接器
Rust 需要某种连接器来操作。它很可能已经安装了，但是当您尝试编译 Rust 问题并得到错误信息告诉您链接器无法执行时，您需要手动安装一个。C 有一个正确的链接器，所以值得按照[原来的教程](https://viewsourcecode.org/snaptoken/kilo/index.html)如何安装 C。

## 检查你的安装
要验证 Rust 是否安装正确，请运行以下命令:

```shell
$ rustc --version
```

运行 cargo 命令：

```shell
$ cargo --version
```

在这两种情况下，您都应该看到程序名称、版本号和一些其他信息。如果没有，请参考官方 Rust Book 的安装章节来解决您的安装问题。

## main 函数
在目录中初始化 rust 项目：

```shell
$ cargo init hecto
```

`hecto` 是我们将要构建的文本编辑器的名称。执行此命令将创建一个名为 hecto 的文件夹，该文件夹已经设置了 git (因此包含一个名为 `.git` 目录 和一个名为 `.gitignore` 的文件)。我们不打算在本教程中使用 git，因此在本教程期间可以忽略这两个文件或文件。

让我们看看还发生了什么:

```
@@ -0,0 +1,9 @@ @@-00 + 19@    
+            [package] [包裹]            
+            name 姓名 = "hecto" “ Hecto”            
+            version 版本 = "0.1.0" “0.1.0”            
+            authors 作家 = ["Philipp Flenker <hello@philippflenker.com>"]            
+            edition 版本 = "2018" “2018”            
+            
+            # See more keys and their definitions at https://doc.rust-lang.org/cargo/reference/manifest.html            
+            
+            [dependencies]
```

```
+            fn  好吧main 总台() {            
+                println!("Hello, world!" “你好，世界!”);            
+            }
```

看看 [Github](https://github.com/pflenker/hecto-tutorial/tree/initialization) 上的这一步

在我们进入细节之前，让我们先弄清楚我们看到的是什么。这是一个差异，它显示了一个代码更改和下一个代码更改之间的变化。在每个步骤的下面，您将在 Github 上找到指向当前步骤的链接，在这里您还可以浏览本教程的其他文件的内容。万一你迷路了，这是非常有用的。

每个 diff 都会告诉您文件是否已被添加、修改或删除，如果需要修改，它会以红色显示已删除的代码，以绿色显示已添加的代码。

正如您可以从 diff 中看到的那样，`cargo init` 已经创建了一个名为 `Cargo.toml` 的文件，其中预先填充了一些信息。更具体地说，方括号表示 section（部分）。下面的依赖项部分为空，package 部分包含:
* 程序的名称，默认为 crate 名称
* 程序的当前版本，默认情况下，是 `0.1.0` 
* 一个作者名单，在这种情况下，预先填写了我的姓名和邮件地址
* 编译器版本。它声明 Rust 应该使用哪个版本编译器

如果您熟悉 JavaScript，那么 Cargo.toml 可以与 package.json 相类似。它描述了您的包以及它对其他包的依赖关系。

此外，还有一个名为 `src/main.rs` 的文件，其中已经包含了一些源代码:

```rust
fn main() {
    println!("Hello, world!");
}
```

这段代码定义了一个名为 main 的函数。main() 函数比较特别。它是运行程序时的默认起始点。当 main() 函数执行完返回时，程序退出并将控制传递回操作系统。因此，在没有任何关于 Rust 的知识的情况下，我们可以推断这个节目遵循了传统的仪式，打印出了“Hello, world!”。看看我们怎么做。

Rust 是一种编译语言。这意味着我们需要通过 Rust 编译器运行程序，将其转换为可执行文件。然后，我们运行该可执行文件，就像在命令行上运行任何其他程序一样。

要编译 `main.rs`，请确保当前目录位于 `hecto` 文件夹中(在 `cargo init hecto` 初始化后执行 `cd hecto`) ，然后在 shell 中运行 `cargo build`。输出结果类似于下方所示:

```
Compiling hecto v0.1.0 (/Users/flenker/Documents/Repositories/hecto)
Finished dev [unoptimized + debuginfo] target(s) in 0.45s
```

这将生成一个名为 `hecto` 的可执行文件，位于名为 `target/debug/` 的文件夹中。此外，还将创建一个名为 `Cargo.lock` 的新文件。它是自动生成的，不应该被手动修改。当我们稍后将依赖项添加到项目中时，`Cargo.lock` 也将被更新。它的更改与本教程无关，因此忽略该文件的差异。

如果进一步查看 `target/debug` 的内容，您会发现生成了很多的文件。这些文件主要是为了提高重建代码的效率(稍后详细介绍)。除了可执行文件本身之外，其他文件在运行程序时都不需要。输入 `./target/debug/hecto` 或 `./hecto` 并按 Enter。程序应输出 Hello，world！然后退出。

## 编译和运行
由于编译和运行程序是非常常见的，因此 Rust 将这两个步骤结合在一起成为 `cargo run`。如果现在在 `cargo build` 之后运行该命令，您可能会注意到输出有一点变化。现在看起来类似于这样:

```
    Finished dev [unoptimized + debuginfo] target(s) in 0.00s
     Running `target/debug/hecto`
Hello, world!
```

正如您看到的那样，cargo 不会像以前那样输出以 Compiling 开始。因为 Rust 告诉我们当前版本的 main.rs 已经被编译过了。如果自上次编译以来没有修改 `main.rs`，那么 RUst 就不会再次进行编译。如果 `main.rs` 发生了更改，则会重新编译 main.rs。这对于需要编译许多不同组件的大型项目很有用，因为当您只更改一个组件的源代码时，大多数组件不应该需要重新编译的。

尝试将 main.rs 中的输出值更改为其他字符串。然后运行 `cargo run`，您应该看到它编译。检查结果，看看是否输出了更改后的字符串。然后将其改回 `Hello，World`，重新编译，并确保返回 `Hello，World`。

在本教程中，大多数步骤，您都需要重新编译代码，查看它是否在您的代码中发现任何错误，然后通过调用 `cargo run` 运行 `hecto`。

在[下一章](https://www.philippflenker.com/hecto-chapter-2/)中，我们将读取用户的按键输入。
