>* Implementing Spinlock for RISC-V OS in Rust 译文（在 Rust 中为 RISC-V 操作系统实现自旋锁） **pending**
>* 原文链接：https://vmm.dev/en/rust/spinlock.md
>* 原文作者：[o8vm](https://github.com/o8vm)
>* 译文来自：https://github.com/suhanyujie/article-transfer-rs/
>* 译者：[suhanyujie](https://github.com/suhanyujie)
>* ps：水平有限，翻译不当之处，还请指正。
>* 标签：Rust，spinlock


Spinlock is one of the most basic synchronization implementations and is one of the first components to consider when implementing an OS.

This article will briefly review the basics of spinlock, how to implement it for self-made OS in Rust, and its advantages over the C language.

* The need for synchronization
* Atomic Operation
    * Test and Set(TAS)
* Simple Spinlock
* A conservative spinlock
* OS Development with Rust

All code examples are written in Rust unless otherwise noted.

## The need for synchronization
When multiple processes run parallel and touch shared data, their exclusion control becomes problematic.
For example, suppose that each process executes its code that modifies the shared data. However, if the code blocks can be performed independently and unrestrictedly, the data is unexpected depending on the timing of the data read/write. This is called a race condition.

The code that causes the race condition is called the critical section. There are synchronization implementations such as spinlock to protect this critical section,

A spinlock uses loops and CPU atomic instructions to limit the number of processes that can execute the critical section to at most one. Note that the name "spin" comes from the looping.

In the Rust language, it is used in the following way.
Here, the spinlock is defined as a data structure called Mutex (MUTual EXclusion).

```rust
let m: Mutex<usize> = Mutex::new(5);
{
    // critical section
    let mut val = m.lock(); // aqcuire lock
    *val = 6;               
}                           // release lock
```

The Mutex wraps and protects the type usize data. Then, the protected data cannot be accessed unless the lock is acquired with lock(). It also waits in a loop if another process has acquired the lock. Then, the lock is released by drop when it goes out of scope.

Let's also look at an example of spinlock in C compared to Rust. In general, C uses a lock management structure or shared variable to manage the state of the lock, and functions such as acquire() and release() are used to acquire and release the lock.

```rust
struct lock vallock;
int m = 5;
aqcuire(&vallock);
// critical section
m = 6;
release(&vallock);
```

In C, there is always the problem of forgetting to release a lock because you have to insert code to acquire and release a lock. And it is also more dangerous than in Rust because you can access the shared data without acquiring the lock.

Now that we have an overview of spinlock let's look at spinlock from the perspective of atomic operation.
