>* How we built a VS Code extension with Rust, WebAssembly, and TypeScript 译文（如何用 Rust，WebAssembly，TypeScript 构建 VS Code 扩展）  **pending**
>* 原文链接：https://www.osohq.com/post/building-vs-code-extension-with-rust-wasm-typescript
>* 原文作者：[Gabe Jackson](https://github.com/osohq/oso)
>* 译文来自：https://github.com/suhanyujie/article-transfer-rs/
>* 译者：[suhanyujie](https://github.com/suhanyujie)
>* 译者博客：[suhanyujie](https://ishenghuo.cnblogs.com/)
>* ps：水平有限，翻译不当之处，还请指正。
>* 标签：Rust，VS Code extension

我们构建了 Oso，这是一个包含电池的框架，用于在您的应用程序中构建授权。Oso 的核心是 [Polar](https://docs.osohq.com/learn/polar-foundations.html)，一种用于编写授权策略的声明性语言。虽然有一天我们会发布一项功能，让全感知的 Oso 为您编写您的授权策略，但在此期间，我们认为改善开发人员编写 Polar 策略的体验是一个好主意。

随着 Oso 的 0.24 版本，我们发布了一个 [VS Code 扩展](https://marketplace.visualstudio.com/items?itemName=osohq.oso)，它可以在 VS Code 中显示策略错误和警告，而不是要求您运行应用程序来查看它们。这缩短了编辑策略的开发周期，让您有更多时间花在任何让您开心的事情上（我们也有更多时间花在确保我们的有意识测试版发布的杀戮命令是一次性故障上）。

在这篇文章中，我们将讨论构建扩展时的设计决策，并在此过程中涉及语言服务器协议，更一般地说，IDE 扩展是如何工作的。如果您曾经想为您选择的语言编写 IDE 扩展，那么这篇文章可能适合您。

## 为什么我们选择 VS Code
我们希望我们的用户在各种环境中编写 Polar，并且那里有各种各样的 IDE。我们将最初的关注点限制在 VS Code 上，因为我们收到了来自社区的最多请求。我们还知道，将来我们会希望将功能扩展到 VS Code 之外，并且在语言服务器协议 (LSP) 之上构建将最大限度地减少将扩展功能移植到 Vim、IntelliJ、Emacs 和其他编辑。

IDE 集成可以提供大量功能，例如自动完成和跳转到定义。对于我们最初的尝试，我们将其范围缩小到显示政策错误和警告。在编写 Polar 策略时，这些可能是解析错误（例如“您缺少分号”）或验证错误（例如“您的策略引用了未知术语”）。尽快向开发人员提出这些问题具有很大的价值。将自己限制在这些诊断上可以改善开发人员的体验，而不会给我们提供无穷无尽的功能列表来实现。

## 使用语言服务器协议 (LSP)
Microsoft 的[开源语言服务器协议](https://microsoft.github.io/language-server-protocol/overviews/lsp/overview/)通过 JSON-RPC 桥抽象了一组通用的 IDE 功能。每个使用 LSP 的 IDE 都会通过该桥发送相同的事件集，例如“文档已打开”或“用户已请求特定符号的所有引用”。每个使用 LSP 的语言服务器都可以实现一次像 document-on-hover这样的功能，并在每个 LSP-fluent 编辑器中获得该功能。

在语言服务器协议之前，如果我们想构建一个 Polar 语言服务器来显示 VS Code 中的错误和警告，我们会使用 VSCode 特定的 API。然后，当我们想将相同的功能移植到 IntelliJ 时，我们必须找出 IntelliJ 中的相应 API，对于 Vim、Emacs 和所有其他 IDE 也是如此。构建相同功能集的 N 个轻微变体将是大量重复的工作。

## 设计
对于最初的设计，我们知道一些事情：

* 我们想用 Rust 编写语言服务器。Polar 解释器是用 Rust 编写的，因此用 Rust 编写语言服务器意味着可以很容易地在我们的 Rust 核心中的现有 API 之上构建。此外，团队非常了解 Rust，它通常是我们在新建项目中的默认设置。
* 语言服务器将是一个新的 crate，`polar-language-server` 以避免 `polar-core` 与特定于 LSP 的代码有关。
* 该扩展将使用 TypeScript 编写，因为 VS Code 期望扩展具有一个 JavaScript 入口点文件，该文件公开 VS Code 的钩子以管理扩展的生命周期。

最后一点提出了我们必须回答的第一个大问题：TypeScript 和 Rust 之间的边界应该在哪里？

## 为什么我们选择 WebAssembly
如果我们想最大化我们用 Rust 编写的数量，一个最小的 TypeScript 扩展将处理扩展激活，在一个单独的进程中启动 Polar 语言服务器（用 Rust 编写），然后，让 VS Code 与服务器通过语言服务器协议通信。但是，这种方式掩盖了我们如何分发语言服务器。VS Code 用户将通过 VS Code 扩展市场安装扩展，但他们如何将 Polar 语言服务器安装到他们的系统上？

如果我们用 Rust 编写服务器，我们有四个主要的分发选项。我们可以快速排除其中两个：

* 将新的 crate 发布到 [crates.io](http://crates.io/) 并让人们 `cargo install` 使用 VS Code 扩展。这很容易排除，因为我们不能要求人们安装兼容的 Rust 工具链与 Cargo 只是为了使用我们的 VS Code 扩展。
* 从 Rust crate 构建特定于平台的二进制文件，或者要求人们手动下载正确的二进制文件，或者做一些主机识别魔法来尝试为他们的平台下载正确的二进制文件。这也很容易排除，因为作为由 100 多个 GitHub Actions 作业组成的发布过程的自豪父母，我们[非常熟悉多平台构建的复杂性](https://www.osohq.com/post/cross-platform-rust-libraries)。

我们有两个可行的选择：

* 利用我们现有的多平台管道并将语言服务器作为可执行文件捆绑到 Oso 语言库中。我们已经为每个库提供了一个可执行的 REPL，所以这里有一些现有技术。
* 将服务器编译为 WebAssembly (Wasm) 并在 Node.js（VS Code 的本机运行时）中运行整个扩展（客户端和服务器）。

上面列表中将语言服务器捆绑到语言库 (3) 和编译到 Wasm (4) 的比较和对比：

* 捆绑意味着语言服务器中的 Oso 版本将始终与项目中安装的版本相匹配，因为语言服务器将来自已安装的库。走 Wasm 路线意味着我们将始终在扩展中发布最新版本的 Oso，并且项目未使用最新版本的人可能会在开发过程中看到与运行时不同的错误和警告，例如，如果我们引入新的验证检查或对解析器进行重大更改。
* 与编译到 Wasm 相比，捆绑需要编写和维护更多的新代码。对于捆绑，我们必须弄清楚如何将新的语言服务器 API 插入到我们为 Python、Node.js、Ruby、Rust、Java 和 Go 构建和分发的现有语言库中，同时编译一个新的 crate 到 Wasm 根本不会影响语言库。
* 捆绑会导致扩展和语言服务器分开分发，所以我们必须考虑如何处理运行不同主要版本的语言服务器协议的两个部分。可以公平地假设这不是一个常见问题，但对于 Wasm 选项，我们仍然不需要考虑这一点，我们将带有扩展名的语言服务器作为附加 `.wasm` 文件分发。
    * 还值得注意的是，语言服务器协议启动握手涉及服务器告诉客户端它支持哪些协议，因此，即使我们分开分发扩展和语言服务器（捆绑选项），我们也可以启用新的客户端功能，而不会破坏与运行相同主要 LSP 版本的旧服务器的向后兼容性。

这两种选择都涉及权衡，我们真的可以选择任何一种方式。然而，我们最终决定使用选项 4——将服务器编译为 WebAssembly 并将客户端和服务器打包为一个 JavaScript 包 —— 因为它看起来实现起来更快，使打包和分发相对简单，而且语言 JavaScript 生态系统中的服务器脚手架包比 Rust 更成熟。

## 决定反对 Rust LSP 库
有三个主要的 Rust crates 为语言服务器提供脚手架：tower-lsp、lspower（ 的一个分支tower-lsp）和lsp-server. 两者都tower-lsp需要 lspower 将语言服务器实现为由 Tower 提供支持的异步服务，Tower 是异步 Rust 的框架。我们认为异步的好处不值得增加语言服务器的复杂性。（值得一提的是，Rust 中最成熟的 LSP 实现 Rust Analyzer 背后的人似乎同意。）首先，扩展在单独的进程中运行，不会影响编辑器的响应能力，并且无论如何解析 Polar 策略并且显示错误和警告是一个非常快速的操作（~1 ms）。此外，还有一些未解决的 问题关于核心模型 tower-lsp 以及 lspowerTower 的请求/响应模型是否自然适合语言服务器协议，语言服务器协议更自然地映射到客户端→服务器和服务器→客户端的一对流。我们还研究 `lsp-server` 了， Rust Analyzer 使用的同步语言服务器脚手架，但缺乏文档和明确的目标，不是“目前是一个好的通用库”让我们望而却步。

最终，我们决定三个脚手架板条箱都不值得进行权衡。我们决定自己编写语言服务器。

## 实现
随着我们的目光投向 Wasm，在 Rust 中实现语言服务器基本上顺利进行，我们最终实现了一个干净的、由三部分组成的设计，其中包括一个 TypeScript 入口点模块，一个第二个微型 TypeScript 模块，它将连接处理委托给微软的 vscode-languageserver 包，并将polar-language-servercrate 编译为 Wasm。

[扩展的入口](https://github.com/osohq/oso/tree/2d5e72fa667f56917524eef8bdcaf903e9ae0fb1/vscode/oso/client)点是一个 TypeScript 模块，它公开 activate() 和 deactivate() 运行，以便 VS Code 可以管理扩展的生命周期。激活后，该扩展会为每个 VS Code 工作区[启动一个单独的语言服务器进程，以支持多根工作区](https://github.com/osohq/oso/blob/2d5e72fa667f56917524eef8bdcaf903e9ae0fb1/vscode/oso/client/src/index.ts#L257)。

在我们为每个工作区启动的流程和 `polar-language-server` Wasm 模块之间有若干层。顶部是 Microsoft vscode-languageclient 包提供的 LanguageClient 类的实例。该 LanguageClient 实例负责启动语言服务器，处理 VS Code 和服务器之间的通信，并将一些配置传递给 VS Code，例如我们对特定工作区感兴趣的文件类型。

当 LanguageClient 启动微型“服务器” TypeScript 模块时，我们初始化 Wasm-ifiedpolar-language-server 并从 TypeScript 管理 LSP 连接的服务器端。通过在 TypeScript 中保留连接管理，我们能够将所有逻辑委托给 vscode-languageserver 包，而不必自己在 polar-language-server. 我们将所有 LSP 通知消息（例如textDocument/didOpen）直接转发到 polar-language-server，该PolarLanguageServer::on_notification 方法将消息从 JSON 反序列化为优秀lsp_typescrate 提供的 Rust 类型。我们还将 JavaScript 回调传递给 polar-language-server 构造函数，以便它可以将诊断信息通过 LSP 分界线发送回 VS Code。

不受我们委派给 Microsoft 的 JavaScript 包的所有繁忙工作的影响，polar-language-server板条箱有一项工作：每当特定工作区的 Polar 策略状态发生变化（即，.polar添加、更改或删除文件）时，它会重新加载策略，翻译任何错误或警告到 VS Code 诊断中，并将这些诊断通过栅栏返回到 VS Code，然后在编辑器中内联显示它们。

## 包起来
通过将语言服务器拆分为一个精简的 TypeScript 包装器和一个 Rust 核心，我们能够在 JavaScript 生态系统中存在的一些不错的 LSP 脚手架之上构建，同时仍然在 Rust 中实现所有新的诊断争论逻辑。随着在 VS Code 中编辑策略，扩展现在会出现问题，从而减少在编辑器和终端之间来回切换的时间，而将更多时间集中在编写授权策略上：

![](https://images.osohq.com/building-vs-code-extension-with-rust-wasm-typescript/Blog%20writing%2012e0d5933594447e9ae0546e49c1645a/Untitled.png)

我们很乐意听到有关新扩展和/或您希望支持的任何功能或编辑器的反馈；请加入我们的 Slack或打开 GitHub 问题。此外，如果这个项目听起来很有趣，可以通过工作获得报酬，我们正在招聘！

